#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include <richedit.h>
#include "StackFunc.h"

#define AEC_FUNCTIONS
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "Resources\Resource.h"


//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackClear
#include "StackFunc.h"

//Defines
#define STRID_SHOW             1
#define STRID_PANEL            2
#define STRID_BACKGROUNDCOLOR  3
#define STRID_TEXTCOLOR        4
#define STRID_BORDERCOLOR      5
#define STRID_BOOKMARKCOLOR    6
#define STRID_BOOKMARKS        7
#define STRID_SETBOOKMARK      8
#define STRID_DELBOOKMARK      10
#define STRID_DELALLBOOKMARK   11
#define STRID_NEXTBOOKMARK     12
#define STRID_PREVBOOKMARK     13
#define STRID_OK               14
#define STRID_CANCEL           15

#define DLLA_LINEBOARD_CALCMAXWIDTH  1
#define DLLA_LINEBOARD_GETBOARDRECT  10

typedef struct _WINDOWBOARD {
  struct _WINDOWBOARD *next;
  struct _WINDOWBOARD *prev;
  HWND hWnd;
  int nBoardWidth;
  BOOL bCalcMaxWidth;
  BOOL bUpdateBookmarks;
  HSTACK hBookmarkStack;
} WINDOWBOARD;

typedef struct _BOOKMARK {
  struct _BOOKMARK *next;
  struct _BOOKMARK *prev;
  int nLine;
  AEPOINT *lpPoint;
} BOOKMARK;

//Functions prototypes
BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetBookmarkProc(void *lpParameter);
BOOL CALLBACK DelBookmarkProc(void *lpParameter);
BOOL CALLBACK DelAllBookmarkProc(void *lpParameter);
BOOL CALLBACK NextBookmarkProc(void *lpParameter);
BOOL CALLBACK PrevBookmarkProc(void *lpParameter);
void StackBeginBoards();
void StackEndBoards(HSTACK *hStack);
void StackUpdateBoards(HSTACK *hStack);
void RecalcMaxWidth(HWND hWnd);
void RecalcMaxWidthEx(HWND hWnd);
void SetEditRect(HWND hWnd, int nNewWidth, int nOldWidth);
WINDOWBOARD* StackInsertBoard(HSTACK *hStack);
WINDOWBOARD* StackGetBoard(HSTACK *hStack, HWND hWnd);
void StackDeleteBoard(HSTACK *hStack, HWND hWnd);
void StackFreeBoards(HSTACK *hStack);
BOOKMARK* StackInsertBookmark(WINDOWBOARD *wb, int nLine);
BOOL StackDeleteBookmark(WINDOWBOARD *wb, int nLine);
int StackGetBookmark(WINDOWBOARD *wb, int nLine);
int StackGetNextBookmark(WINDOWBOARD *wb, int nLine);
int StackGetPrevBookmark(WINDOWBOARD *wb, int nLine);
void StackGetBookmarkRect(WINDOWBOARD *wb, int nLine, RECT *rc);
void StackSaveMobileBookmarks(WINDOWBOARD *wb);
void StackRestoreMobileBookmarks(WINDOWBOARD *wb);
void StackFreeBookmark(WINDOWBOARD *wb);
void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt);
int GetCurrentLine(HWND hWnd);
void SetCurrentLine(HWND hWnd, int nLine);
DWORD ScrollCaret(HWND hWnd);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA();
void SaveOptionsW();
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hMdiClient;
HICON hMainIcon;
BOOL bOldWindows;
BOOL bOldRichEdit;
BOOL bAkelEdit;
BOOL bMDI;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
BOOL bDocumentOpenning=FALSE;
BOOL bShowBoard=TRUE;
DWORD dwStatusPosType=0;
HSTACK hWindowStack={0};
COLORREF crBackground=RGB(0xE8, 0xE4, 0xE0);
COLORREF crText=RGB(0x00, 0x00, 0x00);
COLORREF crBorder=RGB(0x80, 0x80, 0x80);
COLORREF crBookmark=RGB(0xFF, 0xFF, 0xFF);
DWORD dwSetBookmark=3117;    //"Alt+Ins"
DWORD dwDelBookmark=3118;    //"Alt+Del"
DWORD dwDelAllBookmark=3374; //"Alt+Shift+Del"
DWORD dwNextBookmark=3106;   //"Alt+PageDown"
DWORD dwPrevBookmark=3105;   //"Alt+PageUp"
PLUGINFUNCTIONA *pfaSetBookmark=NULL;
PLUGINFUNCTIONA *pfaDelBookmark=NULL;
PLUGINFUNCTIONA *pfaDelAllBookmark=NULL;
PLUGINFUNCTIONA *pfaNextBookmark=NULL;
PLUGINFUNCTIONA *pfaPrevBookmark=NULL;
PLUGINFUNCTIONW *pfwSetBookmark=NULL;
PLUGINFUNCTIONW *pfwDelBookmark=NULL;
PLUGINFUNCTIONW *pfwDelAllBookmark=NULL;
PLUGINFUNCTIONW *pfwNextBookmark=NULL;
PLUGINFUNCTIONW *pfwPrevBookmark=NULL;
WINDOWBOARD *lpCurrentWindowBoard=NULL;
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;
WNDPROCDATA *NewEditProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="LineBoard";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Settings(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (bOldWindows)
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcA);
  else
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcW);

  //If plugin already loaded, stay in memory, but show as non-active
  if (pd->bInMemory) pd->nUnload=UD_NONUNLOAD_NONACTIVE;
}

extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (pd->lParam)
  {
    if (bInitMain)
    {
      if (IsExtCallParamValid(pd->lParam, 1))
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_LINEBOARD_CALCMAXWIDTH)
        {
          RecalcMaxWidthEx(pd->hWndEdit);
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_LINEBOARD_GETBOARDRECT)
        {
          WINDOWBOARD *lpBoard;
          HWND hWnd=NULL;
          RECT *lpRect=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            hWnd=(HWND)GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            lpRect=(RECT *)GetExtCallParam(pd->lParam, 3);

          if (hWnd && lpRect)
          {
            if (lpBoard=StackGetBoard(&hWindowStack, hWnd))
            {
              GetClientRect(hWnd, lpRect);
              lpRect->right=lpBoard->nBoardWidth;
            }
          }
        }
      }

      //Stay in memory, and don't change active status
      pd->nUnload=UD_NONUNLOAD_UNCHANGE;
    }
  }
  else
  {
    //Is plugin already loaded?
    if (bInitMain)
    {
      if (bShowBoard)
      {
        StackEndBoards(&hWindowStack);
      }
      StackFreeBoards(&hWindowStack);
      UninitMain();
    }
    else
    {
      InitMain();

      if (!pd->bOnStart)
      {
        StackBeginBoards();
      }

      //Stay in memory, and show as active
      pd->nUnload=UD_NONUNLOAD_ACTIVE;
    }
  }
}

BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndShowBoard;
  static HWND hWndBackgroundColor;
  static HWND hWndBackgroundColorLabel;
  static HWND hWndTextColor;
  static HWND hWndTextColorLabel;
  static HWND hWndBorderColor;
  static HWND hWndBorderColorLabel;
  static HWND hWndBookmarkColor;
  static HWND hWndBookmarkColorLabel;
  static HWND hWndSetBookmark;
  static HWND hWndDelBookmark;
  static HWND hWndDelAllBookmark;
  static HWND hWndNextBookmark;
  static HWND hWndPrevBookmark;
  static COLORREF crBackgroundDlg;
  static COLORREF crTextDlg;
  static COLORREF crBorderDlg;
  static COLORREF crBookmarkDlg;
  static BOOL bShowBoardDlg;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndShowBoard=GetDlgItem(hDlg, IDC_SHOWBOARD);
    hWndBackgroundColor=GetDlgItem(hDlg, IDC_BACKGROUNDCOLOR_BUTTON);
    hWndBackgroundColorLabel=GetDlgItem(hDlg, IDC_BACKGROUNDCOLOR_LABEL);
    hWndTextColor=GetDlgItem(hDlg, IDC_TEXTCOLOR_BUTTON);
    hWndTextColorLabel=GetDlgItem(hDlg, IDC_TEXTCOLOR_LABEL);
    hWndBorderColor=GetDlgItem(hDlg, IDC_BORDERCOLOR_BUTTON);
    hWndBorderColorLabel=GetDlgItem(hDlg, IDC_BORDERCOLOR_LABEL);
    hWndBookmarkColor=GetDlgItem(hDlg, IDC_BOOKMARKCOLOR_BUTTON);
    hWndBookmarkColorLabel=GetDlgItem(hDlg, IDC_BOOKMARKCOLOR_LABEL);
    hWndSetBookmark=GetDlgItem(hDlg, IDC_SETBOOKMARK);
    hWndDelBookmark=GetDlgItem(hDlg, IDC_DELBOOKMARK);
    hWndDelAllBookmark=GetDlgItem(hDlg, IDC_DELALLBOOKMARK);
    hWndNextBookmark=GetDlgItem(hDlg, IDC_NEXTBOOKMARK);
    hWndPrevBookmark=GetDlgItem(hDlg, IDC_PREVBOOKMARK);

    SetDlgItemTextA(hDlg, IDC_SHOWBOARD, GetLangStringA(wLangSystem, STRID_SHOW));
    SetDlgItemTextA(hDlg, IDC_PANEL_GROUP, GetLangStringA(wLangSystem, STRID_PANEL));
    SetDlgItemTextA(hDlg, IDC_BACKGROUNDCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_BACKGROUNDCOLOR));
    SetDlgItemTextA(hDlg, IDC_TEXTCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_TEXTCOLOR));
    SetDlgItemTextA(hDlg, IDC_BORDERCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_BORDERCOLOR));
    SetDlgItemTextA(hDlg, IDC_BOOKMARKCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_BOOKMARKCOLOR));
    SetDlgItemTextA(hDlg, IDC_BOOKMARK_GROUP, GetLangStringA(wLangSystem, STRID_BOOKMARKS));
    SetDlgItemTextA(hDlg, IDC_SETBOOKMARK_LABEL, GetLangStringA(wLangSystem, STRID_SETBOOKMARK));
    SetDlgItemTextA(hDlg, IDC_DELBOOKMARK_LABEL, GetLangStringA(wLangSystem, STRID_DELBOOKMARK));
    SetDlgItemTextA(hDlg, IDC_DELALLBOOKMARK_LABEL, GetLangStringA(wLangSystem, STRID_DELALLBOOKMARK));
    SetDlgItemTextA(hDlg, IDC_NEXTBOOKMARK_LABEL, GetLangStringA(wLangSystem, STRID_NEXTBOOKMARK));
    SetDlgItemTextA(hDlg, IDC_PREVBOOKMARK_LABEL, GetLangStringA(wLangSystem, STRID_PREVBOOKMARK));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    if (bShowBoard) SendMessage(hWndShowBoard, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndSetBookmark, HKM_SETHOTKEY, dwSetBookmark, 0);
    SendMessage(hWndDelBookmark, HKM_SETHOTKEY, dwDelBookmark, 0);
    SendMessage(hWndDelAllBookmark, HKM_SETHOTKEY, dwDelAllBookmark, 0);
    SendMessage(hWndNextBookmark, HKM_SETHOTKEY, dwNextBookmark, 0);
    SendMessage(hWndPrevBookmark, HKM_SETHOTKEY, dwPrevBookmark, 0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndSetBookmark, pfaSetBookmark?dwSetBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndDelBookmark, pfaDelBookmark?dwDelBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndDelAllBookmark, pfaDelAllBookmark?dwDelAllBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndNextBookmark, pfaNextBookmark?dwNextBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndPrevBookmark, pfaPrevBookmark?dwPrevBookmark:0);

    bShowBoardDlg=bShowBoard;
    crBackgroundDlg=crBackground;
    crTextDlg=crText;
    crBorderDlg=crBorder;
    crBookmarkDlg=crBookmark;

    PostMessage(hDlg, WM_COMMAND, IDC_SHOWBOARD, 0);
  }
  else if (uMsg == WM_DRAWITEM)
  {
    DRAWITEMSTRUCT *dis=(DRAWITEMSTRUCT *)lParam;
    HBRUSH hBrush;
    DWORD dwColor=GetSysColor(COLOR_BTNFACE);

    if (wParam == IDC_BACKGROUNDCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crBackgroundDlg;
    }
    else if (wParam == IDC_TEXTCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crTextDlg;
    }
    else if (wParam == IDC_BORDERCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crBorderDlg;
    }
    else if (wParam == IDC_BOOKMARKCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crBookmarkDlg;
    }

    if (hBrush=CreateSolidBrush(dwColor))
    {
      FillRect(dis->hDC, &dis->rcItem, hBrush);
      if (dis->hwndItem == GetFocus())
        DrawFocusRect(dis->hDC, &dis->rcItem);
      DeleteObject(hBrush);
    }
    return TRUE;
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SHOWBOARD)
    {
      bShowBoardDlg=SendMessage(hWndShowBoard, BM_GETCHECK, 0, 0);
      EnableWindow(hWndBackgroundColor, bShowBoardDlg);
      EnableWindow(hWndBackgroundColorLabel, bShowBoardDlg);
      EnableWindow(hWndTextColor, bShowBoardDlg);
      EnableWindow(hWndTextColorLabel, bShowBoardDlg);
      EnableWindow(hWndBorderColor, bShowBoardDlg);
      EnableWindow(hWndBorderColorLabel, bShowBoardDlg);
      EnableWindow(hWndBookmarkColor, bShowBoardDlg);
      EnableWindow(hWndBookmarkColorLabel, bShowBoardDlg);
      InvalidateRect(hDlg, NULL, FALSE);
    }
    else if (LOWORD(wParam) == IDC_BACKGROUNDCOLOR_BUTTON ||
             LOWORD(wParam) == IDC_TEXTCOLOR_BUTTON ||
             LOWORD(wParam) == IDC_BORDERCOLOR_BUTTON ||
             LOWORD(wParam) == IDC_BOOKMARKCOLOR_BUTTON)
    {
      CHOOSECOLORA ccA={0};
      COLORREF crCustColors[16]={0};
      COLORREF *lpcrColor=NULL;
      HWND hWndButton=NULL;

      if (LOWORD(wParam) == IDC_BACKGROUNDCOLOR_BUTTON)
      {
        lpcrColor=&crBackgroundDlg;
        hWndButton=hWndBackgroundColor;
      }
      else if (LOWORD(wParam) == IDC_TEXTCOLOR_BUTTON)
      {
        lpcrColor=&crTextDlg;
        hWndButton=hWndTextColor;
      }
      else if (LOWORD(wParam) == IDC_BORDERCOLOR_BUTTON)
      {
        lpcrColor=&crBorderDlg;
        hWndButton=hWndBorderColor;
      }
      else if (LOWORD(wParam) == IDC_BOOKMARKCOLOR_BUTTON)
      {
        lpcrColor=&crBookmarkDlg;
        hWndButton=hWndBookmarkColor;
      }
      ccA.lStructSize  =sizeof(CHOOSECOLORA);
      ccA.hwndOwner    =hDlg;
      ccA.lpCustColors =crCustColors;
      ccA.Flags        =CC_FULLOPEN|CC_RGBINIT;
      ccA.rgbResult    =*lpcrColor;

      if (ChooseColorA(&ccA))
      {
        *lpcrColor=ccA.rgbResult;
        InvalidateRect(hWndButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      crBackground=crBackgroundDlg;
      crText=crTextDlg;
      crBorder=crBorderDlg;
      crBookmark=crBookmarkDlg;

      dwSetBookmark=(WORD)SendMessage(hWndSetBookmark, HKM_GETHOTKEY, 0, 0);
      dwDelBookmark=(WORD)SendMessage(hWndDelBookmark, HKM_GETHOTKEY, 0, 0);
      dwDelAllBookmark=(WORD)SendMessage(hWndDelAllBookmark, HKM_GETHOTKEY, 0, 0);
      dwNextBookmark=(WORD)SendMessage(hWndNextBookmark, HKM_GETHOTKEY, 0, 0);
      dwPrevBookmark=(WORD)SendMessage(hWndPrevBookmark, HKM_GETHOTKEY, 0, 0);

      if (bShowBoardDlg != bShowBoard)
      {
        bShowBoard=bShowBoardDlg;

        if (bShowBoard)
          StackBeginBoards();
        else
          StackEndBoards(&hWindowStack);
      }

      SaveOptionsA();
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        UninitMain();
        InitMain();

        if (bShowBoard)
        {
          StackUpdateBoards(&hWindowStack);
        }
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndShowBoard;
  static HWND hWndBackgroundColor;
  static HWND hWndBackgroundColorLabel;
  static HWND hWndTextColor;
  static HWND hWndTextColorLabel;
  static HWND hWndBorderColor;
  static HWND hWndBorderColorLabel;
  static HWND hWndBookmarkColor;
  static HWND hWndBookmarkColorLabel;
  static HWND hWndSetBookmark;
  static HWND hWndDelBookmark;
  static HWND hWndDelAllBookmark;
  static HWND hWndNextBookmark;
  static HWND hWndPrevBookmark;
  static COLORREF crBackgroundDlg;
  static COLORREF crTextDlg;
  static COLORREF crBorderDlg;
  static COLORREF crBookmarkDlg;
  static BOOL bShowBoardDlg;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndShowBoard=GetDlgItem(hDlg, IDC_SHOWBOARD);
    hWndBackgroundColor=GetDlgItem(hDlg, IDC_BACKGROUNDCOLOR_BUTTON);
    hWndBackgroundColorLabel=GetDlgItem(hDlg, IDC_BACKGROUNDCOLOR_LABEL);
    hWndTextColor=GetDlgItem(hDlg, IDC_TEXTCOLOR_BUTTON);
    hWndTextColorLabel=GetDlgItem(hDlg, IDC_TEXTCOLOR_LABEL);
    hWndBorderColor=GetDlgItem(hDlg, IDC_BORDERCOLOR_BUTTON);
    hWndBorderColorLabel=GetDlgItem(hDlg, IDC_BORDERCOLOR_LABEL);
    hWndBookmarkColor=GetDlgItem(hDlg, IDC_BOOKMARKCOLOR_BUTTON);
    hWndBookmarkColorLabel=GetDlgItem(hDlg, IDC_BOOKMARKCOLOR_LABEL);
    hWndSetBookmark=GetDlgItem(hDlg, IDC_SETBOOKMARK);
    hWndDelBookmark=GetDlgItem(hDlg, IDC_DELBOOKMARK);
    hWndDelAllBookmark=GetDlgItem(hDlg, IDC_DELALLBOOKMARK);
    hWndNextBookmark=GetDlgItem(hDlg, IDC_NEXTBOOKMARK);
    hWndPrevBookmark=GetDlgItem(hDlg, IDC_PREVBOOKMARK);

    SetDlgItemTextW(hDlg, IDC_SHOWBOARD, GetLangStringW(wLangSystem, STRID_SHOW));
    SetDlgItemTextW(hDlg, IDC_PANEL_GROUP, GetLangStringW(wLangSystem, STRID_PANEL));
    SetDlgItemTextW(hDlg, IDC_BACKGROUNDCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_BACKGROUNDCOLOR));
    SetDlgItemTextW(hDlg, IDC_TEXTCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_TEXTCOLOR));
    SetDlgItemTextW(hDlg, IDC_BORDERCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_BORDERCOLOR));
    SetDlgItemTextW(hDlg, IDC_BOOKMARKCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_BOOKMARKCOLOR));
    SetDlgItemTextW(hDlg, IDC_BOOKMARK_GROUP, GetLangStringW(wLangSystem, STRID_BOOKMARKS));
    SetDlgItemTextW(hDlg, IDC_SETBOOKMARK_LABEL, GetLangStringW(wLangSystem, STRID_SETBOOKMARK));
    SetDlgItemTextW(hDlg, IDC_DELBOOKMARK_LABEL, GetLangStringW(wLangSystem, STRID_DELBOOKMARK));
    SetDlgItemTextW(hDlg, IDC_DELALLBOOKMARK_LABEL, GetLangStringW(wLangSystem, STRID_DELALLBOOKMARK));
    SetDlgItemTextW(hDlg, IDC_NEXTBOOKMARK_LABEL, GetLangStringW(wLangSystem, STRID_NEXTBOOKMARK));
    SetDlgItemTextW(hDlg, IDC_PREVBOOKMARK_LABEL, GetLangStringW(wLangSystem, STRID_PREVBOOKMARK));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    if (bShowBoard) SendMessage(hWndShowBoard, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndSetBookmark, HKM_SETHOTKEY, dwSetBookmark, 0);
    SendMessage(hWndDelBookmark, HKM_SETHOTKEY, dwDelBookmark, 0);
    SendMessage(hWndDelAllBookmark, HKM_SETHOTKEY, dwDelAllBookmark, 0);
    SendMessage(hWndNextBookmark, HKM_SETHOTKEY, dwNextBookmark, 0);
    SendMessage(hWndPrevBookmark, HKM_SETHOTKEY, dwPrevBookmark, 0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndSetBookmark, pfwSetBookmark?dwSetBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndDelBookmark, pfwDelBookmark?dwDelBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndDelAllBookmark, pfwDelAllBookmark?dwDelAllBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndNextBookmark, pfwNextBookmark?dwNextBookmark:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndPrevBookmark, pfwPrevBookmark?dwPrevBookmark:0);

    bShowBoardDlg=bShowBoard;
    crBackgroundDlg=crBackground;
    crTextDlg=crText;
    crBorderDlg=crBorder;
    crBookmarkDlg=crBookmark;

    PostMessage(hDlg, WM_COMMAND, IDC_SHOWBOARD, 0);
  }
  else if (uMsg == WM_DRAWITEM)
  {
    DRAWITEMSTRUCT *dis=(DRAWITEMSTRUCT *)lParam;
    HBRUSH hBrush;
    DWORD dwColor=GetSysColor(COLOR_BTNFACE);

    if (wParam == IDC_BACKGROUNDCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crBackgroundDlg;
    }
    else if (wParam == IDC_TEXTCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crTextDlg;
    }
    else if (wParam == IDC_BORDERCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crBorderDlg;
    }
    else if (wParam == IDC_BOOKMARKCOLOR_BUTTON)
    {
      if (bShowBoardDlg)
        dwColor=crBookmarkDlg;
    }

    if (hBrush=CreateSolidBrush(dwColor))
    {
      FillRect(dis->hDC, &dis->rcItem, hBrush);
      if (dis->hwndItem == GetFocus())
        DrawFocusRect(dis->hDC, &dis->rcItem);
      DeleteObject(hBrush);
    }
    return TRUE;
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SHOWBOARD)
    {
      bShowBoardDlg=SendMessage(hWndShowBoard, BM_GETCHECK, 0, 0);
      EnableWindow(hWndBackgroundColor, bShowBoardDlg);
      EnableWindow(hWndBackgroundColorLabel, bShowBoardDlg);
      EnableWindow(hWndTextColor, bShowBoardDlg);
      EnableWindow(hWndTextColorLabel, bShowBoardDlg);
      EnableWindow(hWndBorderColor, bShowBoardDlg);
      EnableWindow(hWndBorderColorLabel, bShowBoardDlg);
      EnableWindow(hWndBookmarkColor, bShowBoardDlg);
      EnableWindow(hWndBookmarkColorLabel, bShowBoardDlg);
      InvalidateRect(hDlg, NULL, FALSE);
    }
    else if (LOWORD(wParam) == IDC_BACKGROUNDCOLOR_BUTTON ||
             LOWORD(wParam) == IDC_TEXTCOLOR_BUTTON ||
             LOWORD(wParam) == IDC_BORDERCOLOR_BUTTON ||
             LOWORD(wParam) == IDC_BOOKMARKCOLOR_BUTTON)
    {
      CHOOSECOLORW ccW={0};
      COLORREF crCustColors[16]={0};
      COLORREF *lpcrColor=NULL;
      HWND hWndButton=NULL;

      if (LOWORD(wParam) == IDC_BACKGROUNDCOLOR_BUTTON)
      {
        lpcrColor=&crBackgroundDlg;
        hWndButton=hWndBackgroundColor;
      }
      else if (LOWORD(wParam) == IDC_TEXTCOLOR_BUTTON)
      {
        lpcrColor=&crTextDlg;
        hWndButton=hWndTextColor;
      }
      else if (LOWORD(wParam) == IDC_BORDERCOLOR_BUTTON)
      {
        lpcrColor=&crBorderDlg;
        hWndButton=hWndBorderColor;
      }
      else if (LOWORD(wParam) == IDC_BOOKMARKCOLOR_BUTTON)
      {
        lpcrColor=&crBookmarkDlg;
        hWndButton=hWndBookmarkColor;
      }
      ccW.lStructSize  =sizeof(CHOOSECOLORW);
      ccW.hwndOwner    =hDlg;
      ccW.lpCustColors =crCustColors;
      ccW.Flags        =CC_FULLOPEN|CC_RGBINIT;
      ccW.rgbResult    =*lpcrColor;

      if (ChooseColorW(&ccW))
      {
        *lpcrColor=ccW.rgbResult;
        InvalidateRect(hWndButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      crBackground=crBackgroundDlg;
      crText=crTextDlg;
      crBorder=crBorderDlg;
      crBookmark=crBookmarkDlg;

      dwSetBookmark=(WORD)SendMessage(hWndSetBookmark, HKM_GETHOTKEY, 0, 0);
      dwDelBookmark=(WORD)SendMessage(hWndDelBookmark, HKM_GETHOTKEY, 0, 0);
      dwDelAllBookmark=(WORD)SendMessage(hWndDelAllBookmark, HKM_GETHOTKEY, 0, 0);
      dwNextBookmark=(WORD)SendMessage(hWndNextBookmark, HKM_GETHOTKEY, 0, 0);
      dwPrevBookmark=(WORD)SendMessage(hWndPrevBookmark, HKM_GETHOTKEY, 0, 0);

      if (bShowBoardDlg != bShowBoard)
      {
        bShowBoard=bShowBoardDlg;

        if (bShowBoard)
          StackBeginBoards();
        else
          StackEndBoards(&hWindowStack);
      }

      SaveOptionsW();
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        UninitMain();
        InitMain();

        if (bShowBoard)
        {
          StackUpdateBoards(&hWindowStack);
        }
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (uMsg == AKDN_OPENDOCUMENT_START)
  {
    bDocumentOpenning=TRUE;
  }
  else if (uMsg == AKDN_OPENDOCUMENT_FINISH)
  {
    bDocumentOpenning=FALSE;
    RecalcMaxWidthEx((HWND)wParam);
  }
  else if (uMsg == AKDN_EDIT_ONFINISH)
  {
    if (lpCurrentWindowBoard)
    {
      if (lpCurrentWindowBoard->hWnd == (HWND)wParam)
        lpCurrentWindowBoard=NULL;
    }
    StackDeleteBoard(&hWindowStack, (HWND)wParam);
  }
  if (lResult=EditParentMessages(hWnd, uMsg, wParam, lParam))
    return lResult;

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (lResult=EditParentMessages(hWnd, uMsg, wParam, lParam))
    return lResult;

  //Call next procedure
  return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if (((NMHDR *)lParam)->code == AEN_TEXTCHANGING)
      {
        AENTEXTCHANGE *aentc=(AENTEXTCHANGE *)lParam;
        AECHARRANGE cr;
        WINDOWBOARD *lpBoard;
        BOOL bSaveMobileBookmarks=FALSE;

        if ((aentc->dwType & AETCT_SETTEXT) || (aentc->dwType & AETCT_UNDO) || (aentc->dwType & AETCT_REDO))
        {
          bSaveMobileBookmarks=TRUE;
        }
        else if (!(aentc->aes.dwFlags & AESELT_COLUMNON) && AEC_IndexCompare(&aentc->aes.crSel.ciMin, &aentc->aes.crSel.ciMax))
        {
          SendMessage(aentc->hdr.hwndFrom, AEM_GETINDEX, AEGI_FIRSTCHAR, (LPARAM)&cr.ciMin);
          SendMessage(aentc->hdr.hwndFrom, AEM_GETINDEX, AEGI_LASTCHAR, (LPARAM)&cr.ciMax);

          if (!AEC_IndexCompare(&cr.ciMin, &aentc->aes.crSel.ciMin) &&
              !AEC_IndexCompare(&cr.ciMax, &aentc->aes.crSel.ciMax))
          {
            bSaveMobileBookmarks=TRUE;
          }
        }

        if (bSaveMobileBookmarks)
        {
          HWND hWndMaster;
          HWND hWndEdit;
          int nCloneCount=0;

          if (hWndMaster=(HWND)SendMessage(aentc->hdr.hwndFrom, AEM_GETMASTER, 0, 0))
            hWndEdit=hWndMaster;
          else
            hWndEdit=aentc->hdr.hwndFrom;

          do
          {
            if (lpBoard=StackGetBoard(&hWindowStack, hWndEdit))
            {
              StackSaveMobileBookmarks(lpBoard);
              lpBoard->bUpdateBookmarks=TRUE;
            }
          }
          while (hWndMaster && (hWndEdit=(HWND)SendMessage(hWndMaster, AEM_GETCLONE, nCloneCount++, 0)));
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGED)
      {
        AENTEXTCHANGE *aentc=(AENTEXTCHANGE *)lParam;
        WINDOWBOARD *lpBoard;
        BOOL bRestoreMobileBookmarks=FALSE;

        if (aentc->dwType & AETCT_DELETEALL)
        {
          bRestoreMobileBookmarks=TRUE;
        }

        if (bRestoreMobileBookmarks)
        {
          HWND hWndMaster;
          HWND hWndEdit;
          int nCloneCount=0;

          if (hWndMaster=(HWND)SendMessage(aentc->hdr.hwndFrom, AEM_GETMASTER, 0, 0))
            hWndEdit=hWndMaster;
          else
            hWndEdit=aentc->hdr.hwndFrom;

          do
          {
            if (lpBoard=StackGetBoard(&hWindowStack, hWndEdit))
            {
              if (lpBoard->bUpdateBookmarks)
              {
                lpBoard->bUpdateBookmarks=FALSE;
                StackRestoreMobileBookmarks(lpBoard);
              }
            }
          }
          while (hWndMaster && (hWndEdit=(HWND)SendMessage(hWndMaster, AEM_GETCLONE, nCloneCount++, 0)));
        }
      }
    }
  }
  return FALSE;
}

LRESULT CALLBACK NewEditProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_SETFOCUS)
  {
    WINDOWBOARD *lpBoard;

    if (!lpCurrentWindowBoard || hWnd != lpCurrentWindowBoard->hWnd)
    {
      if (lpBoard=StackGetBoard(&hWindowStack, hWnd))
        lpCurrentWindowBoard=lpBoard;
    }
  }
  else if (uMsg == WM_ERASEBKGND)
  {
    if (bAkelEdit)
    {
      WINDOWBOARD *lpBoard;
      RECT rcBoard;

      if (lpBoard=StackGetBoard(&hWindowStack, hWnd))
      {
        GetClientRect(hWnd, &rcBoard);
        rcBoard.right=lpBoard->nBoardWidth;
        SendMessage(hWnd, AEM_LOCKERASERECT, 0, (LPARAM)&rcBoard);
      }
    }
  }
  else if (uMsg == WM_PAINT)
  {
    WINDOWBOARD *lpBoard;
    RECT rcBoard;
    RECT rcEdit;
    LRESULT lResult=0;
    int nLineCount=0;
    int nLineUnwrapCount=0;
    int nWordWrap=0;
    BOOL bRedrawBoard=FALSE;

    if (!(lpBoard=StackGetBoard(&hWindowStack, hWnd)))
    {
      lpBoard=StackInsertBoard(&hWindowStack);
      lpBoard->hWnd=hWnd;
      lpBoard->nBoardWidth=0;
      lpBoard->bCalcMaxWidth=TRUE;
    }
    if (!lpCurrentWindowBoard) lpCurrentWindowBoard=lpBoard;

    if (bShowBoard)
    {
      if (!bDocumentOpenning)
      {
        GetClientRect(hWnd, &rcBoard);
        rcBoard.right=lpBoard->nBoardWidth;

        nLineCount=SendMessage(hWnd, EM_GETLINECOUNT, 0, 0);
        nLineUnwrapCount=nLineCount;
        if (bAkelEdit && !(dwStatusPosType & SPT_LINEWRAP))
        {
          nWordWrap=SendMessage(hWnd, AEM_GETWORDWRAP, 0, (LPARAM)NULL);
          if (nWordWrap)
            nLineUnwrapCount=SendMessage(hWnd, AEM_GETUNWRAPLINE, nLineCount - 1, 0) + 1;
        }
        bRedrawBoard=TRUE;
      }

      lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

      if (bRedrawBoard)
      {
        AEFOLD *lpFold=NULL;
        char szLine[MAX_PATH];
        AECHARINDEX ciFirstLine;
        AECHARINDEX ciLastLine;
        POINT ptFirstChar;
        TEXTMETRICA tm;
        HDC hDC;
        HRGN hRgn;
        HRGN hRgnOld;
        HFONT hFont;
        HFONT hFontOld;
        HBRUSH hBrush;
        HBRUSH hBookmark;
        HBRUSH hBookmarkOld;
        HPEN hPen;
        HPEN hPenOld;
        RECT rcDraw;
        RECT rcBrush;
        POINT ptTextOut;
        SIZE sizeLine;
        int nLineLen;
        int nOldBoardWidth;
        int nCharHeight;
        int nCharAverageWidth;
        int nFirstChar;
        int nLastChar;
        int nUnwrappedLine=0;
        int bWrapped;

        if (hDC=GetDC(hWnd))
        {
          hRgn=CreateRectRgn(rcBoard.left, rcBoard.top, rcBoard.right, rcBoard.bottom);
          hRgnOld=(HRGN)SelectObject(hDC, hRgn);
          hFont=CreateFontIndirectA((LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
          hFontOld=(HFONT)SelectObject(hDC, hFont);
          hBrush=CreateSolidBrush(crBackground);
          hBookmark=CreateSolidBrush(crBookmark);
          SetBkColor(hDC, crBackground);

          //Get char height
          if (bAkelEdit)
          {
            nCharHeight=SendMessage(hWnd, AEM_GETCHARSIZE, AECS_HEIGHT, 0);
          }
          else
          {
            GetTextMetricsA(hDC, &tm);
            nCharHeight=tm.tmHeight;
          }

          if (lpBoard->nBoardWidth)
          {
            if (hPen=CreatePen(PS_SOLID, 0, crBorder))
            {
              hPenOld=(HPEN)SelectObject(hDC, hPen);
              MoveToEx(hDC, rcBoard.right - 1, rcBoard.top, NULL);
              LineTo(hDC, rcBoard.right - 1, rcBoard.bottom);
              if (hPenOld) SelectObject(hDC, hPenOld);
              DeleteObject(hPen);
            }
          }

          if (lpBoard->bCalcMaxWidth)
          {
            lpBoard->bCalcMaxWidth=FALSE;
            nLineLen=wsprintfA(szLine, "%d", nLineUnwrapCount);
            if (GetTextExtentPoint32A(hDC, szLine, nLineLen, &sizeLine))
            {
              nCharAverageWidth=sizeLine.cx / nLineLen;
              nOldBoardWidth=lpBoard->nBoardWidth;
              lpBoard->nBoardWidth=sizeLine.cx + nCharAverageWidth;
              SetEditRect(hWnd, lpBoard->nBoardWidth, nOldBoardWidth);
            }
          }
          else
          {
            if (bAkelEdit)
            {
              SendMessage(hWnd, AEM_GETINDEX, AEGI_FIRSTVISIBLELINE, (LPARAM)&ciFirstLine);
              SendMessage(hWnd, AEM_GETINDEX, AEGI_LASTVISIBLELINE, (LPARAM)&ciLastLine);
              ptFirstChar.x=0;
              ptFirstChar.y=SendMessage(hWnd, AEM_VPOSFROMLINE, AECT_CLIENT, ciFirstLine.nLine);
            }
            else
            {
              SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcEdit);
              nFirstChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcEdit.left);
              nLastChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcEdit.right);
              ciFirstLine.nLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstChar);
              ciLastLine.nLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nLastChar);
              GetPosFromChar(hWnd, nFirstChar, &ptFirstChar);
            }
            rcDraw.left=0;
            rcDraw.top=ptFirstChar.y;
            rcDraw.right=rcBoard.right - 1;
            rcDraw.bottom=rcBoard.bottom;

            while (ciFirstLine.nLine <= ciLastLine.nLine && rcDraw.top < rcDraw.bottom)
            {
              if (!SendMessage(hWnd, AEM_ISLINECOLLAPSED, (WPARAM)&lpFold, ciFirstLine.nLine))
              {
                if (nWordWrap)
                  nUnwrappedLine=SendMessage(hWnd, AEM_GETUNWRAPLINE, ciFirstLine.nLine, 0);
                else
                  nUnwrappedLine=ciFirstLine.nLine;
                nLineLen=wsprintfA(szLine, "%d", nUnwrappedLine + 1);
                if (!GetTextExtentPoint32A(hDC, szLine, nLineLen, &sizeLine)) break;
                nCharAverageWidth=sizeLine.cx / nLineLen;

                if (sizeLine.cx + nCharAverageWidth > rcBoard.right)
                {
                  nOldBoardWidth=lpBoard->nBoardWidth;
                  lpBoard->nBoardWidth=sizeLine.cx + nCharAverageWidth;
                  SetEditRect(hWnd, lpBoard->nBoardWidth, nOldBoardWidth);
                  break;
                }
                ptTextOut.x=rcBoard.right - sizeLine.cx - (nCharAverageWidth / 2);
                ptTextOut.y=rcDraw.top;
                bWrapped=FALSE;

                if (nWordWrap)
                {
                  if (ciFirstLine.lpLine->prev && ciFirstLine.lpLine->prev->nLineBreak == AELB_WRAP)
                  {
                    //Erase line space
                    rcBrush.left=0;
                    rcBrush.right=rcDraw.right;
                    rcBrush.top=rcDraw.top;
                    rcBrush.bottom=rcDraw.top + nCharHeight;
                    FillRect(hDC, &rcBrush, hBrush);
                    bWrapped=TRUE;
                  }
                }

                if (StackGetBookmark(lpBoard, ciFirstLine.nLine) != -1)
                {
                  if (!bWrapped)
                  {
                    //Erase line space
                    rcBrush.left=0;
                    rcBrush.right=rcDraw.right;
                    rcBrush.top=rcDraw.top;
                    rcBrush.bottom=rcDraw.top + nCharHeight;
                    FillRect(hDC, &rcBrush, hBrush);
                  }

                  //Draw bookmark
                  hBookmarkOld=(HBRUSH)SelectObject(hDC, hBookmark);
                  RoundRect(hDC, 0, rcDraw.top, rcDraw.right, rcDraw.top + nCharHeight, nCharHeight / 3, nCharHeight / 3);
                  if (hBookmarkOld) SelectObject(hDC, hBookmarkOld);

                  if (!bWrapped)
                  {
                    //Draw number
                    SetBkMode(hDC, TRANSPARENT);
                    SetTextColor(hDC, crText);
                    TextOutA(hDC, ptTextOut.x, ptTextOut.y, szLine, nLineLen);
                    SetBkMode(hDC, OPAQUE);
                  }
                }
                else
                {
                  if (!bWrapped)
                  {
                    //Erase space before number
                    rcBrush.left=0;
                    rcBrush.right=ptTextOut.x;
                    rcBrush.top=rcDraw.top;
                    rcBrush.bottom=rcDraw.top + nCharHeight;
                    FillRect(hDC, &rcBrush, hBrush);

                    //Draw number
                    SetTextColor(hDC, crText);
                    rcBrush.left=ptTextOut.x;
                    rcBrush.right=ptTextOut.x + sizeLine.cx;
                    ExtTextOutA(hDC, ptTextOut.x, ptTextOut.y, ETO_OPAQUE, &rcBrush, szLine, nLineLen, NULL);

                    //Erase space after number
                    rcBrush.left=ptTextOut.x + sizeLine.cx;
                    rcBrush.right=rcDraw.right;
                    FillRect(hDC, &rcBrush, hBrush);
                  }
                }
                rcDraw.top+=nCharHeight;
              }
              if (bAkelEdit)
                ciFirstLine.lpLine=ciFirstLine.lpLine->next;
              ++ciFirstLine.nLine;
            }

            if (ciFirstLine.nLine > ciLastLine.nLine && rcDraw.top < rcDraw.bottom)
            {
              FillRect(hDC, &rcDraw, hBrush);
            }
          }
          if (hBookmark) DeleteObject(hBookmark);
          if (hBrush) DeleteObject(hBrush);
          if (hFontOld) SelectObject(hDC, hFontOld);
          DeleteObject(hFont);
          if (hRgnOld) SelectObject(hDC, hRgnOld);
          DeleteObject(hRgn);
          ReleaseDC(hWnd, hDC);
        }
      }
    }
    if (bShowBoard) return lResult;
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    POINT pt;

    if (bShowBoard)
    {
      if (lpCurrentWindowBoard)
      {
        pt.x=LOWORD(lParam);
        pt.y=HIWORD(lParam);

        if (pt.x < lpCurrentWindowBoard->nBoardWidth)
        {
          BOOKMARK *lpBookmark;
          RECT rc={0, 0, 1, 1};
          RECT rcBookmark;
          int nClickChar;
          int nClickLine;

          nClickChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&pt);
          nClickLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nClickChar);

          if (StackGetBookmark(lpCurrentWindowBoard, nClickLine) >= 0)
          {
            if (StackDeleteBookmark(lpCurrentWindowBoard, nClickLine))
            {
              StackGetBookmarkRect(lpCurrentWindowBoard, nClickLine, &rcBookmark);
              InvalidateRect(hWnd, &rcBookmark, FALSE);
            }
          }
          else
          {
            if (lpBookmark=StackInsertBookmark(lpCurrentWindowBoard, nClickLine))
            {
              InvalidateRect(hWnd, &rc, FALSE);
            }
          }
          return 0;
        }
      }
    }
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewEditProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_SETFOCUS)
  {
    WINDOWBOARD *lpBoard;

    if (!lpCurrentWindowBoard || hWnd != lpCurrentWindowBoard->hWnd)
    {
      if (lpBoard=StackGetBoard(&hWindowStack, hWnd))
        lpCurrentWindowBoard=lpBoard;
    }
  }
  else if (uMsg == WM_ERASEBKGND)
  {
    if (bAkelEdit)
    {
      WINDOWBOARD *lpBoard;
      RECT rcBoard;

      if (lpBoard=StackGetBoard(&hWindowStack, hWnd))
      {
        GetClientRect(hWnd, &rcBoard);
        rcBoard.right=lpBoard->nBoardWidth;
        SendMessage(hWnd, AEM_LOCKERASERECT, 0, (LPARAM)&rcBoard);
      }
    }
  }
  else if (uMsg == WM_PAINT)
  {
    WINDOWBOARD *lpBoard;
    RECT rcBoard;
    RECT rcEdit;
    LRESULT lResult=0;
    int nLineCount=0;
    int nLineUnwrapCount=0;
    int nWordWrap=0;
    BOOL bRedrawBoard=FALSE;

    if (!(lpBoard=StackGetBoard(&hWindowStack, hWnd)))
    {
      lpBoard=StackInsertBoard(&hWindowStack);
      lpBoard->hWnd=hWnd;
      lpBoard->nBoardWidth=0;
      lpBoard->bCalcMaxWidth=TRUE;
    }
    if (!lpCurrentWindowBoard) lpCurrentWindowBoard=lpBoard;

    if (bShowBoard)
    {
      if (!bDocumentOpenning)
      {
        GetClientRect(hWnd, &rcBoard);
        rcBoard.right=lpBoard->nBoardWidth;

        nLineCount=SendMessage(hWnd, EM_GETLINECOUNT, 0, 0);
        nLineUnwrapCount=nLineCount;
        if (bAkelEdit && !(dwStatusPosType & SPT_LINEWRAP))
        {
          nWordWrap=SendMessage(hWnd, AEM_GETWORDWRAP, 0, (LPARAM)NULL);
          if (nWordWrap)
            nLineUnwrapCount=SendMessage(hWnd, AEM_GETUNWRAPLINE, nLineCount - 1, 0) + 1;
        }
        bRedrawBoard=TRUE;
      }

      lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

      if (bRedrawBoard)
      {
        AEFOLD *lpFold=NULL;
        wchar_t wszLine[MAX_PATH];
        AECHARINDEX ciFirstLine;
        AECHARINDEX ciLastLine;
        POINT ptFirstChar;
        TEXTMETRICW tm;
        HDC hDC;
        HRGN hRgn;
        HRGN hRgnOld;
        HFONT hFont;
        HFONT hFontOld;
        HBRUSH hBrush;
        HBRUSH hBookmark;
        HBRUSH hBookmarkOld;
        HPEN hPen;
        HPEN hPenOld;
        RECT rcDraw;
        RECT rcBrush;
        POINT ptTextOut;
        SIZE sizeLine;
        int nLineLen;
        int nOldBoardWidth;
        int nCharHeight;
        int nCharAverageWidth;
        int nFirstChar;
        int nLastChar;
        int nUnwrappedLine=0;
        int bWrapped;

        if (hDC=GetDC(hWnd))
        {
          hRgn=CreateRectRgn(rcBoard.left, rcBoard.top, rcBoard.right, rcBoard.bottom);
          hRgnOld=(HRGN)SelectObject(hDC, hRgn);
          hFont=CreateFontIndirectW((LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
          hFontOld=(HFONT)SelectObject(hDC, hFont);
          hBrush=CreateSolidBrush(crBackground);
          hBookmark=CreateSolidBrush(crBookmark);
          SetBkColor(hDC, crBackground);

          //Get char height
          if (bAkelEdit)
          {
            nCharHeight=SendMessage(hWnd, AEM_GETCHARSIZE, AECS_HEIGHT, 0);
          }
          else
          {
            GetTextMetricsW(hDC, &tm);
            nCharHeight=tm.tmHeight;
          }

          if (lpBoard->nBoardWidth)
          {
            if (hPen=CreatePen(PS_SOLID, 0, crBorder))
            {
              hPenOld=(HPEN)SelectObject(hDC, hPen);
              MoveToEx(hDC, rcBoard.right - 1, rcBoard.top, NULL);
              LineTo(hDC, rcBoard.right - 1, rcBoard.bottom);
              if (hPenOld) SelectObject(hDC, hPenOld);
              DeleteObject(hPen);
            }
          }

          if (lpBoard->bCalcMaxWidth)
          {
            lpBoard->bCalcMaxWidth=FALSE;
            nLineLen=wsprintfW(wszLine, L"%d", nLineUnwrapCount);
            if (GetTextExtentPoint32W(hDC, wszLine, nLineLen, &sizeLine))
            {
              nCharAverageWidth=sizeLine.cx / nLineLen;
              nOldBoardWidth=lpBoard->nBoardWidth;
              lpBoard->nBoardWidth=sizeLine.cx + nCharAverageWidth;
              SetEditRect(hWnd, lpBoard->nBoardWidth, nOldBoardWidth);
            }
          }
          else
          {
            if (bAkelEdit)
            {
              SendMessage(hWnd, AEM_GETINDEX, AEGI_FIRSTVISIBLELINE, (LPARAM)&ciFirstLine);
              SendMessage(hWnd, AEM_GETINDEX, AEGI_LASTVISIBLELINE, (LPARAM)&ciLastLine);
              ptFirstChar.x=0;
              ptFirstChar.y=SendMessage(hWnd, AEM_VPOSFROMLINE, AECT_CLIENT, ciFirstLine.nLine);
            }
            else
            {
              SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcEdit);
              nFirstChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcEdit.left);
              nLastChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcEdit.right);
              ciFirstLine.nLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstChar);
              ciLastLine.nLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nLastChar);
              GetPosFromChar(hWnd, nFirstChar, &ptFirstChar);
            }
            rcDraw.left=0;
            rcDraw.top=ptFirstChar.y;
            rcDraw.right=rcBoard.right - 1;
            rcDraw.bottom=rcBoard.bottom;

            while (ciFirstLine.nLine <= ciLastLine.nLine && rcDraw.top < rcDraw.bottom)
            {
              if (!SendMessage(hWnd, AEM_ISLINECOLLAPSED, (WPARAM)&lpFold, ciFirstLine.nLine))
              {
                if (nWordWrap)
                  nUnwrappedLine=SendMessage(hWnd, AEM_GETUNWRAPLINE, ciFirstLine.nLine, 0);
                else
                  nUnwrappedLine=ciFirstLine.nLine;
                nLineLen=wsprintfW(wszLine, L"%d", nUnwrappedLine + 1);
                if (!GetTextExtentPoint32W(hDC, wszLine, nLineLen, &sizeLine)) break;
                nCharAverageWidth=sizeLine.cx / nLineLen;

                if (sizeLine.cx + nCharAverageWidth > rcBoard.right)
                {
                  nOldBoardWidth=lpBoard->nBoardWidth;
                  lpBoard->nBoardWidth=sizeLine.cx + nCharAverageWidth;
                  SetEditRect(hWnd, lpBoard->nBoardWidth, nOldBoardWidth);
                  break;
                }
                ptTextOut.x=rcBoard.right - sizeLine.cx - (nCharAverageWidth / 2);
                ptTextOut.y=rcDraw.top;
                bWrapped=FALSE;

                if (nWordWrap)
                {
                  if (ciFirstLine.lpLine->prev && ciFirstLine.lpLine->prev->nLineBreak == AELB_WRAP)
                  {
                    //Erase line space
                    rcBrush.left=0;
                    rcBrush.right=rcDraw.right;
                    rcBrush.top=rcDraw.top;
                    rcBrush.bottom=rcDraw.top + nCharHeight;
                    FillRect(hDC, &rcBrush, hBrush);
                    bWrapped=TRUE;
                  }
                }

                if (StackGetBookmark(lpBoard, ciFirstLine.nLine) != -1)
                {
                  if (!bWrapped)
                  {
                    //Erase line space
                    rcBrush.left=0;
                    rcBrush.right=rcDraw.right;
                    rcBrush.top=rcDraw.top;
                    rcBrush.bottom=rcDraw.top + nCharHeight;
                    FillRect(hDC, &rcBrush, hBrush);
                  }

                  //Draw bookmark
                  hBookmarkOld=(HBRUSH)SelectObject(hDC, hBookmark);
                  RoundRect(hDC, 0, rcDraw.top, rcDraw.right, rcDraw.top + nCharHeight, nCharHeight / 3, nCharHeight / 3);
                  if (hBookmarkOld) SelectObject(hDC, hBookmarkOld);

                  if (!bWrapped)
                  {
                    //Draw number
                    SetBkMode(hDC, TRANSPARENT);
                    SetTextColor(hDC, crText);
                    TextOutW(hDC, ptTextOut.x, ptTextOut.y, wszLine, nLineLen);
                    SetBkMode(hDC, OPAQUE);
                  }
                }
                else
                {
                  if (!bWrapped)
                  {
                    //Erase space before number
                    rcBrush.left=0;
                    rcBrush.right=ptTextOut.x;
                    rcBrush.top=rcDraw.top;
                    rcBrush.bottom=rcDraw.top + nCharHeight;
                    FillRect(hDC, &rcBrush, hBrush);

                    //Draw number
                    SetTextColor(hDC, crText);
                    rcBrush.left=ptTextOut.x;
                    rcBrush.right=ptTextOut.x + sizeLine.cx;
                    ExtTextOutW(hDC, ptTextOut.x, ptTextOut.y, ETO_OPAQUE, &rcBrush, wszLine, nLineLen, NULL);

                    //Erase space after number
                    rcBrush.left=ptTextOut.x + sizeLine.cx;
                    rcBrush.right=rcDraw.right;
                    FillRect(hDC, &rcBrush, hBrush);
                  }
                }
                rcDraw.top+=nCharHeight;
              }
              if (bAkelEdit)
                ciFirstLine.lpLine=ciFirstLine.lpLine->next;
              ++ciFirstLine.nLine;
            }

            if (ciFirstLine.nLine > ciLastLine.nLine && rcDraw.top < rcDraw.bottom)
            {
              FillRect(hDC, &rcDraw, hBrush);
            }
          }
          if (hBookmark) DeleteObject(hBookmark);
          if (hBrush) DeleteObject(hBrush);
          if (hFontOld) SelectObject(hDC, hFontOld);
          DeleteObject(hFont);
          if (hRgnOld) SelectObject(hDC, hRgnOld);
          DeleteObject(hRgn);
          ReleaseDC(hWnd, hDC);
        }
      }
    }
    if (bShowBoard) return lResult;
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    POINT pt;

    if (lpCurrentWindowBoard)
    {
      if (bShowBoard)
      {
        pt.x=LOWORD(lParam);
        pt.y=HIWORD(lParam);

        if (pt.x < lpCurrentWindowBoard->nBoardWidth)
        {
          BOOKMARK *lpBookmark;
          RECT rc={0, 0, 1, 1};
          RECT rcBookmark;
          int nClickChar;
          int nClickLine;

          nClickChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&pt);
          nClickLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nClickChar);

          if (StackGetBookmark(lpCurrentWindowBoard, nClickLine) >= 0)
          {
            if (StackDeleteBookmark(lpCurrentWindowBoard, nClickLine))
            {
              StackGetBookmarkRect(lpCurrentWindowBoard, nClickLine, &rcBookmark);
              InvalidateRect(hWnd, &rcBookmark, FALSE);
            }
          }
          else
          {
            if (lpBookmark=StackInsertBookmark(lpCurrentWindowBoard, nClickLine))
            {
              InvalidateRect(hWnd, &rc, FALSE);
            }
          }
          return 0;
        }
      }
    }
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

BOOL CALLBACK SetBookmarkProc(void *lpParameter)
{
  BOOKMARK *lpBookmark;
  RECT rc={0, 0, 1, 1};
  int nCurrentLine;

  if (lpCurrentWindowBoard)
  {
    nCurrentLine=GetCurrentLine(lpCurrentWindowBoard->hWnd);

    if (lpBookmark=StackInsertBookmark(lpCurrentWindowBoard, nCurrentLine))
    {
      if (bShowBoard) InvalidateRect(lpCurrentWindowBoard->hWnd, &rc, FALSE);
    }
  }
  return TRUE;
}

BOOL CALLBACK DelBookmarkProc(void *lpParameter)
{
  RECT rcBookmark;
  int nCurrentLine;

  if (lpCurrentWindowBoard)
  {
    nCurrentLine=GetCurrentLine(lpCurrentWindowBoard->hWnd);

    if (StackDeleteBookmark(lpCurrentWindowBoard, nCurrentLine))
    {
      StackGetBookmarkRect(lpCurrentWindowBoard, nCurrentLine, &rcBookmark);
      if (bShowBoard) InvalidateRect(lpCurrentWindowBoard->hWnd, &rcBookmark, FALSE);
    }
  }
  return TRUE;
}

BOOL CALLBACK DelAllBookmarkProc(void *lpParameter)
{
  RECT rcBoard;

  if (lpCurrentWindowBoard)
  {
    StackFreeBookmark(lpCurrentWindowBoard);

    if (bShowBoard)
    {
      GetClientRect(lpCurrentWindowBoard->hWnd, &rcBoard);
      rcBoard.right=lpCurrentWindowBoard->nBoardWidth;
      InvalidateRect(lpCurrentWindowBoard->hWnd, &rcBoard, FALSE);
    }
  }
  return TRUE;
}

BOOL CALLBACK NextBookmarkProc(void *lpParameter)
{
  int nBookmarkLine;

  if (lpCurrentWindowBoard)
  {
    if ((nBookmarkLine=StackGetNextBookmark(lpCurrentWindowBoard, GetCurrentLine(lpCurrentWindowBoard->hWnd))) >= 0)
    {
      SetCurrentLine(lpCurrentWindowBoard->hWnd, nBookmarkLine);
    }
  }
  return TRUE;
}

BOOL CALLBACK PrevBookmarkProc(void *lpParameter)
{
  int nBookmarkLine;

  if (lpCurrentWindowBoard)
  {
    if ((nBookmarkLine=StackGetPrevBookmark(lpCurrentWindowBoard, GetCurrentLine(lpCurrentWindowBoard->hWnd))) >= 0)
    {
      SetCurrentLine(lpCurrentWindowBoard->hWnd, nBookmarkLine);
    }
  }
  return TRUE;
}

void StackBeginBoards()
{
  HWND hWndFrame;
  HWND hWndChild;
  RECT rc={0, 0, 1, 1};

  if (!bMDI)
  {
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
        InvalidateRect(hWndChild, &rc, FALSE);

      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }
  }
  else
  {
    hWndFrame=GetWindow(hMdiClient, GW_CHILD);

    while (hWndFrame)
    {
      hWndChild=GetWindow(hWndFrame, GW_CHILD);

      while (hWndChild)
      {
        if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
          InvalidateRect(hWndChild, &rc, FALSE);

        hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
      }
      hWndFrame=GetNextWindow(hWndFrame, GW_HWNDNEXT);
    }
  }
}

void StackEndBoards(HSTACK *hStack)
{
  WINDOWBOARD *lpBoard=(WINDOWBOARD *)hStack->first;
  int nOldWidth;

  while (lpBoard)
  {
    if (IsWindow(lpBoard->hWnd))
    {
      nOldWidth=lpBoard->nBoardWidth;
      lpBoard->nBoardWidth=0;
      SetEditRect(lpBoard->hWnd, lpBoard->nBoardWidth, nOldWidth);

      lpBoard->bCalcMaxWidth=TRUE;
      InvalidateRect(lpBoard->hWnd, NULL, TRUE);
    }
    lpBoard=lpBoard->next;
  }
}

void StackUpdateBoards(HSTACK *hStack)
{
  WINDOWBOARD *lpBoard=(WINDOWBOARD *)hStack->first;
  RECT rcBoard;

  while (lpBoard)
  {
    GetClientRect(lpBoard->hWnd, &rcBoard);
    rcBoard.right=lpBoard->nBoardWidth;
    InvalidateRect(lpBoard->hWnd, &rcBoard, FALSE);

    lpBoard=lpBoard->next;
  }
}

void RecalcMaxWidth(HWND hWnd)
{
  WINDOWBOARD *lpBoard;
  RECT rc={0, 0, 1, 1};

  if (lpBoard=StackGetBoard(&hWindowStack, hWnd))
  {
    lpBoard->bCalcMaxWidth=TRUE;
    InvalidateRect(hWnd, &rc, FALSE);
  }
}

void RecalcMaxWidthEx(HWND hWnd)
{
  HWND hWndMaster;
  HWND hWndClone;
  int i;

  if (hWndMaster=(HWND)SendMessage(hWnd, AEM_GETMASTER, 0, 0))
  {
    RecalcMaxWidth(hWndMaster);

    for (i=0; hWndClone=(HWND)SendMessage(hWndMaster, AEM_GETCLONE, i, 0); ++i)
    {
      RecalcMaxWidth(hWndClone);
    }
  }
  else RecalcMaxWidth(hWnd);
}

void SetEditRect(HWND hWnd, int nNewWidth, int nOldWidth)
{
  RECT rcClient;
  RECT rcDraw;
  int nBorderX;
  int nBorderY;

  GetClientRect(hWnd, &rcClient);
  SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcDraw);

  rcClient.left=nNewWidth + (rcDraw.left - nOldWidth);
  rcClient.right-=(rcClient.right - rcDraw.right);

  if (bOldRichEdit)
  {
    rcClient.left+=3;
    rcClient.top+=4;
    rcClient.right-=3;
    rcClient.bottom-=4;
  }
  else
  {
    if ((GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_CLIENTEDGE) ||
        (GetWindowLong(hWnd, GWL_STYLE) & ES_SUNKEN))
    {
      nBorderX=GetSystemMetrics(SM_CXBORDER);
      nBorderY=GetSystemMetrics(SM_CYBORDER);

      rcClient.left+=nBorderX;
      rcClient.top+=nBorderY;
      rcClient.right-=nBorderX;
      rcClient.bottom-=nBorderY;
    }
  }

  SendMessage(hWnd, EM_SETRECT, 0, (LPARAM)&rcClient);
}

WINDOWBOARD* StackInsertBoard(HSTACK *hStack)
{
  WINDOWBOARD *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(WINDOWBOARD));
  return lpElement;
}

WINDOWBOARD* StackGetBoard(HSTACK *hStack, HWND hWnd)
{
  WINDOWBOARD *lpBoard=(WINDOWBOARD *)hStack->first;

  while (lpBoard)
  {
    if (lpBoard->hWnd == hWnd)
      return lpBoard;

    lpBoard=lpBoard->next;
  }
  return NULL;
}

void StackDeleteBoard(HSTACK *hStack, HWND hWnd)
{
  WINDOWBOARD *lpBoard;

  if (lpBoard=StackGetBoard(hStack, hWnd))
  {
    StackFreeBookmark(lpBoard);
    StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpBoard);
  }
}

void StackFreeBoards(HSTACK *hStack)
{
  WINDOWBOARD *lpBoard=(WINDOWBOARD *)hStack->first;

  while (lpBoard)
  {
    StackFreeBookmark(lpBoard);

    lpBoard=lpBoard->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

BOOKMARK* StackInsertBookmark(WINDOWBOARD *wb, int nLine)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;
  BOOKMARK *lpElement=NULL;

  while (lpBookmark)
  {
    if (!bAkelEdit)
    {
      if (lpBookmark->nLine == nLine)
        return NULL;
      if (lpBookmark->nLine > nLine)
        break;
    }
    else
    {
      if (lpBookmark->lpPoint->ciPoint.nLine == nLine)
        return NULL;
      if (lpBookmark->lpPoint->ciPoint.nLine > nLine)
        break;
    }
    lpBookmark=lpBookmark->next;
  }

  if (!StackInsertBefore((stack **)&wb->hBookmarkStack.first, (stack **)&wb->hBookmarkStack.last, (stack *)lpBookmark, (stack **)&lpElement, sizeof(BOOKMARK)))
  {
    lpElement->nLine=nLine;

    if (bAkelEdit)
    {
      AEPOINT point={0};

      if (SendMessage(wb->hWnd, AEM_GETLINEINDEX, (WPARAM)nLine, (LPARAM)&point.ciPoint))
      {
        point.nPointOffset=AEPTO_IGNORE;
        if (point.ciPoint.lpLine->prev && point.ciPoint.lpLine->prev->nLineBreak == AELB_WRAP)
          point.ciPoint.nCharInLine=min(point.ciPoint.nCharInLine + 1, point.ciPoint.lpLine->nLineLen);
        lpElement->lpPoint=(AEPOINT *)SendMessage(wb->hWnd, AEM_ADDPOINT, 0, (LPARAM)&point);
      }
    }
  }
  return lpElement;
}

BOOL StackDeleteBookmark(WINDOWBOARD *wb, int nLine)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;

  while (lpBookmark)
  {
    if (!bAkelEdit)
    {
      if (lpBookmark->nLine == nLine)
      {
        StackDelete((stack **)&wb->hBookmarkStack.first, (stack **)&wb->hBookmarkStack.last, (stack *)lpBookmark);
        return TRUE;
      }
    }
    else
    {
      if (lpBookmark->lpPoint->ciPoint.nLine == nLine)
      {
        SendMessage(wb->hWnd, AEM_DELPOINT, (WPARAM)lpBookmark->lpPoint, 0);
        StackDelete((stack **)&wb->hBookmarkStack.first, (stack **)&wb->hBookmarkStack.last, (stack *)lpBookmark);
        return TRUE;
      }
    }
    lpBookmark=lpBookmark->next;
  }
  return FALSE;
}

int StackGetBookmark(WINDOWBOARD *wb, int nLine)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;

  while (lpBookmark)
  {
    if (!bAkelEdit)
    {
      if (lpBookmark->nLine > nLine)
        break;
      if (lpBookmark->nLine == nLine)
        return lpBookmark->nLine;
    }
    else
    {
      if (lpBookmark->lpPoint->ciPoint.nLine > nLine)
        break;
      if (lpBookmark->lpPoint->ciPoint.nLine == nLine)
        return lpBookmark->lpPoint->ciPoint.nLine;
    }
    lpBookmark=lpBookmark->next;
  }
  return -1;
}

int StackGetNextBookmark(WINDOWBOARD *wb, int nLine)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;

  while (lpBookmark)
  {
    if (!bAkelEdit)
    {
      if (lpBookmark->nLine > nLine)
        return lpBookmark->nLine;
    }
    else
    {
      if (lpBookmark->lpPoint->ciPoint.nLine > nLine)
        return lpBookmark->lpPoint->ciPoint.nLine;
    }
    lpBookmark=lpBookmark->next;
  }
  return -1;
}

int StackGetPrevBookmark(WINDOWBOARD *wb, int nLine)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;
  BOOKMARK *prev=NULL;

  while (lpBookmark)
  {
    if (!bAkelEdit)
    {
      if (lpBookmark->nLine < nLine)
        prev=lpBookmark;
      else
        break;
    }
    else
    {
      if (lpBookmark->lpPoint->ciPoint.nLine < nLine)
        prev=lpBookmark;
      else
        break;
    }
    lpBookmark=lpBookmark->next;
  }

  if (prev)
  {
    if (!bAkelEdit)
      return prev->nLine;
    else
      return prev->lpPoint->ciPoint.nLine;
  }
  return -1;
}

void StackGetBookmarkRect(WINDOWBOARD *wb, int nLine, RECT *rc)
{
  RECT rcEdit;
  POINT ptTop;
  POINT ptBottom;
  int nLineIndex;

  GetClientRect(wb->hWnd, &rcEdit);

  if ((nLineIndex=SendMessage(wb->hWnd, EM_LINEINDEX, nLine, 0)) >= 0)
  {
    GetPosFromChar(wb->hWnd, nLineIndex, &ptTop);

    if ((nLineIndex=SendMessage(wb->hWnd, EM_LINEINDEX, nLine + 1, 0)) >= 0)
    {
      GetPosFromChar(wb->hWnd, nLineIndex, &ptBottom);

      rc->left=rcEdit.left;
      rc->top=ptTop.y;
      rc->right=wb->nBoardWidth;
      rc->bottom=ptBottom.y;
    }
    else
    {
      rc->left=rcEdit.left;
      rc->top=ptTop.y;
      rc->right=wb->nBoardWidth;
      rc->bottom=rcEdit.bottom;
    }
  }
  else
  {
    rc->left=rcEdit.left;
    rc->top=rcEdit.top;
    rc->right=wb->nBoardWidth;
    rc->bottom=rcEdit.bottom;
  }
}

void StackSaveMobileBookmarks(WINDOWBOARD *wb)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;

  while (lpBookmark)
  {
    lpBookmark->nLine=lpBookmark->lpPoint->ciPoint.nLine;

    lpBookmark=lpBookmark->next;
  }
}

void StackRestoreMobileBookmarks(WINDOWBOARD *wb)
{
  BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;
  BOOKMARK *lpPrevBookmark;

  while (lpBookmark)
  {
    lpPrevBookmark=lpBookmark->next;

    lpBookmark->lpPoint->ciPoint.nLine=lpBookmark->nLine;
    SendMessage(wb->hWnd, AEM_INDEXUPDATE, 0, (LPARAM)&lpBookmark->lpPoint->ciPoint);
    if (lpBookmark->lpPoint->ciPoint.nLine == lpBookmark->nLine)
      lpBookmark->lpPoint->ciPoint.nCharInLine=min(lpBookmark->lpPoint->ciPoint.nCharInLine, lpBookmark->lpPoint->ciPoint.lpLine->nLineLen);
    else
      StackDelete((stack **)&wb->hBookmarkStack.first, (stack **)&wb->hBookmarkStack.last, (stack *)lpBookmark);

    lpBookmark=lpPrevBookmark;
  }
}

void StackFreeBookmark(WINDOWBOARD *wb)
{
  if (bAkelEdit)
  {
    BOOKMARK *lpBookmark=(BOOKMARK *)wb->hBookmarkStack.first;

    while (lpBookmark)
    {
      SendMessage(wb->hWnd, AEM_DELPOINT, (WPARAM)lpBookmark->lpPoint, 0);

      lpBookmark=lpBookmark->next;
    }
  }
  StackClear((stack **)&wb->hBookmarkStack.first, (stack **)&wb->hBookmarkStack.last);
}

void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt)
{
  DWORD dwPosition;

  if (bOldRichEdit)
  {
    dwPosition=SendMessage(hWnd, EM_POSFROMCHAR, nCharIndex, (LPARAM)NULL);
    pt->x=LOWORD(dwPosition);
    pt->y=HIWORD(dwPosition);
  }
  else
  {
    SendMessage(hWnd, EM_POSFROMCHAR, (WPARAM)pt, nCharIndex);
  }
}

int GetCurrentLine(HWND hWnd)
{
  int nLineIndex;

  nLineIndex=SendMessage(hWnd, EM_LINEINDEX, (WPARAM)-1, 0);
  return SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nLineIndex);
}

void SetCurrentLine(HWND hWnd, int nLine)
{
  CHARRANGE cr;
  int nLineIndex;

  if ((nLineIndex=SendMessage(hWnd, EM_LINEINDEX, nLine, 0)) != -1)
  {
    cr.cpMin=nLineIndex;
    cr.cpMax=nLineIndex;

    if (!bAkelEdit)
    {
      SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
    }
    else
    {
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
      ScrollCaret(hWnd);
    }
  }
}

DWORD ScrollCaret(HWND hWnd)
{
  AESCROLLTOPOINT stp;
  DWORD dwScrollResult;

  //Test scroll to caret
  stp.dwFlags=AESC_TEST|AESC_POINTCARET|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
  stp.nOffsetX=1;
  stp.nOffsetY=0;
  dwScrollResult=SendMessage(hWnd, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);

  //Scroll to caret
  stp.dwFlags=AESC_POINTCARET;
  stp.nOffsetX=3;
  stp.nOffsetY=2;
  if (dwScrollResult & AECSE_SCROLLEDX)
    stp.dwFlags|=AESC_OFFSETRECTDIVX;
  if (dwScrollResult & AECSE_SCROLLEDY)
    stp.dwFlags|=AESC_OFFSETRECTDIVY;
  return SendMessage(hWnd, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="ShowBoard";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bShowBoard;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="BackgroundColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="TextColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crText;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="BorderColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBorder;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="BookmarkColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="SetBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwSetBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="DelBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="DelAllBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelAllBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NextBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNextBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="PrevBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwPrevBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"ShowBoard";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bShowBoard;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"BackgroundColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"TextColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crText;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"BorderColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBorder;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"BookmarkColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"SetBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwSetBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"DelBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"DelAllBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelAllBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NextBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNextBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"PrevBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwPrevBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    po.pOptionName="ShowBoard";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bShowBoard;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="BackgroundColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="TextColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crText;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="BorderColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBorder;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="BookmarkColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="SetBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwSetBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="DelBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="DelAllBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelAllBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NextBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNextBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="PrevBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwPrevBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"ShowBoard";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bShowBoard;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"BackgroundColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"TextColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crText;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"BorderColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBorder;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"BookmarkColor";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"SetBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwSetBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"DelBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"DelAllBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwDelAllBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NextBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNextBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"PrevBookmark";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwPrevBookmark;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_SHOW)
      return L"\x041F\x043E\x043A\x0430\x0437\x0430\x0442\x044C";
    if (nStringID == STRID_PANEL)
      return L"\x041F\x0430\x043D\x0435\x043B\x044C";
    if (nStringID == STRID_BACKGROUNDCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0444\x043E\x043D\x0430";
    if (nStringID == STRID_TEXTCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0442\x0435\x043A\x0441\x0442\x0430";
    if (nStringID == STRID_BORDERCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0433\x0440\x0430\x043D\x0438\x0446\x044B";
    if (nStringID == STRID_BOOKMARKCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0437\x0430\x043A\x043B\x0430\x0434\x043A\x0438";
    if (nStringID == STRID_BOOKMARKS)
      return L"\x0417\x0430\x043A\x043B\x0430\x0434\x043A\x0438";
    if (nStringID == STRID_SETBOOKMARK)
      return L"\x0423\x0441\x0442\x0430\x043D\x043E\x0432\x0438\x0442\x044C";
    if (nStringID == STRID_DELBOOKMARK)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C";
    if (nStringID == STRID_DELALLBOOKMARK)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C\x0020\x0432\x0441\x0435";
    if (nStringID == STRID_NEXTBOOKMARK)
      return L"\x0421\x043B\x0435\x0434\x0443\x044E\x0449\x0430\x044F";
    if (nStringID == STRID_PREVBOOKMARK)
      return L"\x041F\x0440\x0435\x0434\x044B\x0434\x0443\x0449\x0430\x044F";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_SHOW)
      return L"Show board";
    if (nStringID == STRID_PANEL)
      return L"Panel";
    if (nStringID == STRID_BACKGROUNDCOLOR)
      return L"Background color";
    if (nStringID == STRID_TEXTCOLOR)
      return L"Text color";
    if (nStringID == STRID_BORDERCOLOR)
      return L"Border color";
    if (nStringID == STRID_BOOKMARKCOLOR)
      return L"Bookmark color";
    if (nStringID == STRID_BOOKMARKS)
      return L"Bookmarks";
    if (nStringID == STRID_SETBOOKMARK)
      return L"Set";
    if (nStringID == STRID_DELBOOKMARK)
      return L"Delete";
    if (nStringID == STRID_DELALLBOOKMARK)
      return L"Delete all";
    if (nStringID == STRID_NEXTBOOKMARK)
      return L"Next";
    if (nStringID == STRID_PREVBOOKMARK)
      return L"Previous";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hMdiClient=pd->hMdiClient;
  hMainIcon=pd->hMainIcon;
  bOldWindows=pd->bOldWindows;
  bOldRichEdit=pd->bOldRichEdit;
  bAkelEdit=pd->bAkelEdit;
  bMDI=pd->bMDI;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
    ReadOptionsA();
  else
    ReadOptionsW();

  dwStatusPosType=SendMessage(hMainWnd, AKD_GETSTATUSPOSTYPE, 0, 0);
}

void InitMain()
{
  bInitMain=TRUE;

  //Add hotkeys
  if (bOldWindows)
  {
    PLUGINFUNCTIONA pf={0};

    if (dwSetBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwSetBookmark))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "LineBoard::Main::SetBookmark");
      pf.wHotkey=(WORD)dwSetBookmark;
      pf.PluginProc=SetBookmarkProc;
      pfaSetBookmark=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwDelBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwDelBookmark))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "LineBoard::Main::DelBookmark");
      pf.wHotkey=(WORD)dwDelBookmark;
      pf.PluginProc=DelBookmarkProc;
      pfaDelBookmark=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwDelAllBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwDelAllBookmark))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "LineBoard::Main::DelAllBookmark");
      pf.wHotkey=(WORD)dwDelAllBookmark;
      pf.PluginProc=DelAllBookmarkProc;
      pfaDelAllBookmark=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwNextBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwNextBookmark))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "LineBoard::Main::NextBookmark");
      pf.wHotkey=(WORD)dwNextBookmark;
      pf.PluginProc=NextBookmarkProc;
      pfaNextBookmark=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwPrevBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwPrevBookmark))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "LineBoard::Main::PrevBookmark");
      pf.wHotkey=(WORD)dwPrevBookmark;
      pf.PluginProc=PrevBookmarkProc;
      pfaPrevBookmark=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
  }
  else
  {
    PLUGINFUNCTIONW pf={0};

    if (dwSetBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwSetBookmark))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"LineBoard::Main::SetBookmark");
      pf.wHotkey=(WORD)dwSetBookmark;
      pf.PluginProc=SetBookmarkProc;
      pfwSetBookmark=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwDelBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwDelBookmark))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"LineBoard::Main::DelBookmark");
      pf.wHotkey=(WORD)dwDelBookmark;
      pf.PluginProc=DelBookmarkProc;
      pfwDelBookmark=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwDelAllBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwDelAllBookmark))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"LineBoard::Main::DelAllBookmark");
      pf.wHotkey=(WORD)dwDelAllBookmark;
      pf.PluginProc=DelAllBookmarkProc;
      pfwDelAllBookmark=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwNextBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwNextBookmark))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"LineBoard::Main::NextBookmark");
      pf.wHotkey=(WORD)dwNextBookmark;
      pf.PluginProc=NextBookmarkProc;
      pfwNextBookmark=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwPrevBookmark && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwPrevBookmark))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"LineBoard::Main::PrevBookmark");
      pf.wHotkey=(WORD)dwPrevBookmark;
      pf.PluginProc=PrevBookmarkProc;
      pfwPrevBookmark=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
  }

  //Subclass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewFrameProc, (LPARAM)&NewFrameProcData);
  }

  if (bOldWindows)
  {
    NewEditProcData=NULL;
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProcA, (LPARAM)&NewEditProcData);
  }
  else
  {
    NewEditProcData=NULL;
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProcW, (LPARAM)&NewEditProcData);
  }
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove hotkeys
  if (bOldWindows)
  {
    if (pfaSetBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaSetBookmark);
      pfaSetBookmark=NULL;
    }
    if (pfaDelBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaDelBookmark);
      pfaDelBookmark=NULL;
    }
    if (pfaDelAllBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaDelAllBookmark);
      pfaDelAllBookmark=NULL;
    }
    if (pfaNextBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaNextBookmark);
      pfaNextBookmark=NULL;
    }
    if (pfaPrevBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaPrevBookmark);
      pfaPrevBookmark=NULL;
    }
  }
  else
  {
    if (pfwSetBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwSetBookmark);
      pfwSetBookmark=NULL;
    }
    if (pfwDelBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwDelBookmark);
      pfwDelBookmark=NULL;
    }
    if (pfwDelAllBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwDelAllBookmark);
      pfwDelAllBookmark=NULL;
    }
    if (pfwNextBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwNextBookmark);
      pfwNextBookmark=NULL;
    }
    if (pfwPrevBookmark)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwPrevBookmark);
      pfwPrevBookmark=NULL;
    }
  }

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
  if (NewFrameProcData)
  {
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
    NewFrameProcData=NULL;
  }
  if (NewEditProcData)
  {
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NULL, (LPARAM)&NewEditProcData);
    NewEditProcData=NULL;
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain)
    {
      if (bShowBoard)
      {
//        StackEndBoards(&hWindowStack);
      }
      StackFreeBoards(&hWindowStack);
      UninitMain();
    }
  }
  return TRUE;
}
